import atexit
from contextlib import contextmanager
import dataclasses
from typing import Any, Generator, Optional, Type
import datetime
import time
import subprocess
import requests
import chili


URL_TEMPL = "http://{hostname}:{port}/jsonrpc"


class CSException(Exception):
    pass


class UnityCommsFn:
    def __init__(self, unity_comms: 'UnityComms', method_name: str):
        self.unity_comms = unity_comms
        self.method_name = method_name

    def __call__(self, ResultClass: Optional[Type] = None, retry: bool = True, **kwargs: Any) -> Any:
        return self.unity_comms.rpc_call(
            method=self.method_name, retry=retry, params_dict=kwargs, ResultClass=ResultClass)


class UnityComms:
    def __init__(
        self,
        port: int,
        server_executable_path: Optional[str] = None,
        logfile: Optional[str] = None,
        hostname: str = "localhost"
    ) -> None:
        """
        :param port: int The port that Unity will run on. Always mandatory. If server_executable_path is provided, we
            will start the server with commandline `--port {port}`
        :param server_executable_path: Optional[str] Path to dedicatd server executable to run
        :param logfile: Optional[str] Logfile to write to. Optional
        :param hostname: str  If providing server_executable_path, must be 'localhost', otherwise the hostname where
            the Unity process is running. Reminder that all network communications are insecure, so best to set this
            to localhost. (For secure network communciations you could use e.g. ssh tunnels)
        """
        self.server_executable_path = server_executable_path
        if server_executable_path is not None:
            assert hostname == "localhost", "Must use hostname localhost if passing in server_executable_path"
        self.hostname = hostname
        self.port = port
        self.session = requests.Session()
        self.jsonrpc_id = 0
        self.logfile = logfile

        if server_executable_path is not None:
            atexit.register(self._kill_server)
            self._start_server()

    def _start_server(self) -> None:
        assert self.server_executable_path is not None
        cmd_line = [self.server_executable_path, '--port', str(self.port)]
        print(cmd_line)
        subprocess.Popen(cmd_line, )

    def _kill_server(self, *args: Any, **kwargs: Any) -> None:
        print('Asking unity to die')
        self.rpc_call('shutdownUnity', retry=False)

    def _rpc_request_dict(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        res = {"method": method, "params": params, "jsonrpc": "2.0", "id": self.jsonrpc_id}
        self.jsonrpc_id += 1
        return res

    @contextmanager
    def blocking_listen(self) -> Generator:
        """
        Uses blocking listens for commands from the python side. Note that on a dedicated unity server, this is
        automatic: no need to call this. This only has an affect when running against a non-dedicated Unity, typically
        the Unity Editor.

        To avoid blocking leaving the Editor blocked after the script exits, this should be used from a `with` block,
        like:

        with unity_comms.blocking_listen():
            # do other stuff here

        (This means we automatically attempt to remove the blocking listen from the Unity Editor, at the end)
        """
        self.rpc_call("setBlockingListen", {"blocking": True})
        try:
            yield None
        finally:
            self.rpc_call("setBlockingListen", {"blocking": False}, retry=False)

    def set_autosimulation(self, auto_simulation: bool) -> None:
        """
        You should call this to turn off autosimulation prior to running any reinforcement learning
        algorithm. Otherwise your environment risks stepping in between your reinforcement learning
        steps :P
        """
        self.rpc_call("setAutosimulation", {"autosimulation": auto_simulation})

    def get_autosimulation(self) -> bool:
        return self.rpc_call("getAutosimulation")

    def __getattr__(self, method_name: str) -> UnityCommsFn:
        if method_name.startswith('_'):
            raise AttributeError()
        return UnityCommsFn(unity_comms=self, method_name=method_name)

    def __getitem__(self, method_name: str) -> UnityCommsFn:
        return UnityCommsFn(unity_comms=self, method_name=method_name)

    def rpc_call(
        self,
        method: str,
        params_dict: Optional[dict[str, Any]] = None,
        ResultClass: Optional[Type] = None,
        retry: bool = True,
        **kwargs: dict[str, Any]
    ) -> Any:
        """
        :param unity_port: int The Port that our Unity application is listening on
        :param method: str The json rpc method we want to call. Usually this is the name of the method
        :param params_dict: Optional[dict[str, Any]] Any parameters we want to pass into the method,
            keyed by parameter name
            Can provide structured data using dataclasses
        :param ResultClass: Optional[Type] If not None, then the returned json dict will be converted
            into this type, using Chili. Should be a dataclass. If None, then the raw result dict will
            be returned instead.
        :param **kwargs: dict[str, Any]  You can also simply pass in parameters by name
        """
        params_dict = params_dict if params_dict else {}
        params_dict.update(kwargs)
        new_dict = {}
        for k, v in params_dict.items():
            if dataclasses.is_dataclass(v):
                v = chili.asdict(v)
            new_dict[k] = v
        params_dict = new_dict
        payload = self._rpc_request_dict(method, params_dict)
        url = URL_TEMPL.format(port=self.port, hostname=self.hostname)
        response = None
        while response is None:
            try:
                res = self.session.post(url, json=payload)
                if res.content is None:
                    print("content is None => skipping")
                    continue
                if res.content == "".encode("utf-8"):
                    # no json content detected => retry
                    continue
                if res.content.decode("utf-8").strip() == "":
                    print("stripped content is empty string => skipping")
                    continue
                res_d = res.json()
                if "error" in res_d:
                    print("res_d", res_d)
                    err_data = res_d["error"]["data"]
                    if isinstance(err_data, dict):
                        print(err_data["ClassName"])
                        print(err_data["Message"])
                        print(err_data["StackTraceString"].replace("\\n", "\n"))
                        raise CSException(err_data["Message"])
                    else:
                        raise CSException(err_data)
                if ResultClass is None:
                    return res_d["result"]
                response = chili.init_dataclass(res_d["result"], ResultClass)
            except requests.exceptions.ConnectionError:
                if retry:
                    print("requests.exceptions.ConnectionError => ignoring, retrying")
                    time.sleep(0.1)
                else:
                    return
            except CSException as e:
                print("payload", payload)
                print("res", res)
                raise e
            except Exception as e:
                print("payload", payload)
                print("res", res)
                print("res.content", res.content)
                print("e", e)
                if self.logfile is not None:
                    with open(self.logfile, 'a') as f:
                        datetime_str = datetime.datetime.now().strftime("%Y%m%d %H%M%S")
                        f.write(f"{datetime_str}: payload {payload}\n")
                        f.write(f"{datetime_str}: res {res}\n")
                        f.write(f"{datetime_str}: res.content {str(res.content)}\n")
                        f.write(f"{datetime_str}: e {e}\n")
                time.sleep(0.1)
        return response



from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray


@dataclass
class RayResults:
    rayDistances: list[list[float]]
    rayHitObjectTypes: list[list[int]]
    NumObjectTypes: int


def ray_results_to_feature_np(
    ray_results: RayResults,
) -> NDArray[np.float32]:
    """
    Takes in the ray results, i.e. hit distances and object types,
    and return a single numpy array, of same shape as each of distances and object types,
    but with an additional 0th dimension of length object_types

    Conceptually, if a number is zero, that ray didnt hit that tag type. If it is near zero,
    it hit that tag type, but it was far away. If it is non-zero and near 1, then the ray hit
    that tag type, and the object was relatively near. So:
    - 0 => tag not seen
    - near 0 => tag far away
    - near 1 => tag nearby

    The number is not upper-bounded: we simply take the inverse of the distance. However, you can
    always apply your own sigmoid or similar, to the output of this function, to bound it.

    In detail: the distances will be inverted, so that nearer gives a higher number, and further gives
    smaller. this number will be assigned to the output plane indexed by object type.
    If object type is -1, the output will be set to 0 across all output channels
    """
    distances_np = np.array(ray_results.rayDistances)
    distances_np = 1 / distances_np
    object_types_np = np.array(ray_results.rayHitObjectTypes)
    # add one feature plane for the object type of -1
    _obs = np.zeros(
        (ray_results.NumObjectTypes + 1, *distances_np.shape), dtype=np.float32
    )
    np.put_along_axis(
        _obs,
        np.expand_dims(object_types_np, axis=0),
        np.expand_dims(distances_np, axis=0),
        axis=0,
    )
    # truncate off the not found object feature plane
    _obs = _obs[:-1]
    return _obs



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['peaceful_pie']

package_data = \
{'': ['*']}

install_requires = \
['chili>=1.8.0,<2.0.0', 'numpy>=1.24.1,<2.0.0', 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'peaceful-pie',
    'version': '1.5.0',
    'description': 'Control Unity from Python!',
    'long_description': '# Peaceful Pie\n\nConnect Python with Unity for reinforcement learning!\n\n# CI\n\n[![CircleCI](https://dl.circleci.com/status-badge/img/gh/hughperkins/peaceful-pie/tree/main.svg?style=svg)](https://dl.circleci.com/status-badge/redirect/gh/hughperkins/peaceful-pie/tree/main)\n\n# Examples\n\n- see [examples](examples)\n\n# Installation\n\nYou need to install both into Unity project, and into your Python environment.\n\n## In Unity\n\n- In Unity, go to Window | Package Manager\n    - change the dropdown in top left to \'Packages: In Project\'\n    - check that Newtonsoft Json appears\n    - if Newtonsoft Json is not already in your project:\n        - click on the \'+\' in the top left, and choose \'Add package by name...\'\n        - put the name `com.unity.nuget.newtonsoft-json`, then click \'Add\'\n- In Unity, in your project\'s "Assets" folder, create a "Plugins" folder, if it doesn\'t already exist\n- First install AustinHarris.JsonRPC:\n    - Download https://www.nuget.org/api/v2/package/AustinHarris.JsonRpc/1.2.3\n    - rename to have filename suffix be `.zip` (you might need to turn on options to see all file extensions)\n    - unzip the resulting zip file\n    - copy `lib/netstandard2.1/AustinHarris.JsonRpc.dll` into your `Plugins` folder\n    - select the file, in your Plugins, and in \'Inspector\' unselect \'validate references\', and click \'Apply\'\n- Click on \'Releases\', on the right of the github page\n    - in the most recent release, download `PeacefulPie.dll`\n- Copy `PeacefulPie.dll` into your Unity project\'s `Plugins` folder\n- If on Mac silicon:\n    - in Unity, goto Plugins, click on `PeacefulPie.dll`\n    - in Inspector, change \'CPU\' setting from \'Intel\' to \'Any CPU\', then click \'Apply\'\n    - do the same for `AustinHarris.JsonRpc.dll`\n\nYou should be good to go :)\n\n## In Python\n\n```\npip install -U peaceful-pie\n```\n\n# Requirements\n\n- currently tested with:\n    - python 3.10\n    - Unity 2021.3.16.f1\n- please create an issue if your preferred platform is not supported (I\'m guessing I might need to downgrade Python a little? Let me know!)\n\n# Dev\n\nI\'m usng Visual Studio for Mac to write this.\n',
    'author': 'Hugh Perkins',
    'author_email': 'hughperkins@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/hughperkins/peaceful-pie',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
